@@ -1,3 +1,5 @@
+import multiprocessing
+from tqdm import tqdm
 import numpy as np
 import pandas as pd
 import click
@@ -5,6 +7,7 @@ from sklearn import preprocessing
 import statsmodels.api as sm
 import sys
 import os
+import numpy as np
 
 from lxml import etree
 import itertools
@@ -20,44 +23,73 @@ except ImportError:
     plt = None
 
 class uniprot:
-    def __init__(self, uniprotfile):
+    def __init__(self, uniprotfile, cache):
         self.namespaces = {'uniprot': "http://uniprot.org/uniprot"}
+        self.cache = cache
         self.df = self.read(uniprotfile)
-
+    
     def read(self, uniprotfile):
+        if uniprotfile.endswith("xml.gz"):
+            cache_filename = uniprotfile.strip("xml.gz") + "parquet"
+        elif uniprotfile.endswith("xml"):
+            cache_filename = uniprotfile.strip("xml") + "parquet"
+        
+        if self.cache and os.path.exists(cache_filename):
+            print("Found cached table")
+            return pd.read_parquet(cache_filename)
+        
         def _extract(lst):
             if len(lst) >= 1:
                 return lst[0]
             else:
                 return None
 
-        df = pd.DataFrame(columns=["protein_id", "protein_name", "ensembl_id", "protein_mw"])
+        df = pd.DataFrame(columns=["protein_id", "protein_name", "gene", "ensembl_id", "protein_mw"])
 
         root = etree.parse(uniprotfile)
 
-        accessions = root.xpath('//uniprot:entry/uniprot:accession/text()', namespaces = self.namespaces)
-        names = root.xpath('//uniprot:entry/uniprot:name/text()', namespaces = self.namespaces)
         mw = root.xpath('//uniprot:entry/uniprot:sequence/@mass', namespaces = self.namespaces)
-        ensembl = root.xpath('//uniprot:entry/uniprot:dbReference[@type="Ensembl"]/uniprot:property[@type="protein sequence ID"]/@value', namespaces = self.namespaces)
+        if root.xpath('//uniprot:entry/uniprot:organism/uniprot:dbReference[@type="NCBI Taxonomy"]/@id', namespaces = self.namespaces)[0] == '559292':
+            ensembl = root.xpath('//uniprot:entry/uniprot:gene/uniprot:name[@type = "ordered locus"]/text()', namespaces = self.namespaces)
+            ensembl_path = './uniprot:gene/uniprot:name[@type = "ordered locus"]/text()'
+        else:
+            ensembl = root.xpath('//uniprot:entry/uniprot:dbReference[@type="Ensembl"]/uniprot:property[@type="protein sequence ID"]/@value', namespaces = self.namespaces)
+            ensembl_path = './uniprot:dbReference[@type="Ensembl"]/uniprot:property[@type="protein sequence ID"]/@value'
 
-        for entry in root.xpath('//uniprot:entry', namespaces = self.namespaces):
+        # TODO: Need to make this parsing more memory efficient
+        entries = root.xpath('//uniprot:entry', namespaces = self.namespaces)
+        rows = []
+        for entry in tqdm(entries, total=len(entries)):
             accession = entry.xpath('./uniprot:accession/text()', namespaces = self.namespaces)
             name = entry.xpath('./uniprot:name/text()', namespaces = self.namespaces)
+            gene = entry.xpath('./uniprot:gene/uniprot:name/text()', namespaces = self.namespaces)
             mw = entry.xpath('./uniprot:sequence/@mass', namespaces = self.namespaces)
-            ensembl = entry.xpath('./uniprot:dbReference[@type="Ensembl"]/uniprot:property[@type="protein sequence ID"]/@value', namespaces = self.namespaces)
-
-            df = df.append({'protein_id': _extract(accession), 'protein_name': _extract(name), 'ensembl_id': ensembl, 'protein_mw': float(_extract(mw))}, ignore_index=True)
+            ensembl = entry.xpath(ensembl_path, namespaces = self.namespaces)
 
+            row = pd.Series({
+                'protein_id': _extract(accession), 
+                'protein_name': _extract(name), 
+                'gene': _extract(gene), 
+                'ensembl_id': ensembl, 
+                'protein_mw': float(_extract(mw))
+            })
+            rows.append(row)
+        
+        # Append each Series object as a new row to df
+        df = pd.concat([df, *[row.to_frame().T for row in rows]], ignore_index=True)
         return df
 
     def to_df(self):
-        return self.df[['protein_id','protein_name','protein_mw']]
+        return self.df[['protein_id','protein_name', 'gene', 'protein_mw']]
 
     def expand(self):
-        ensembl = self.df.apply(lambda x: pd.Series(x['ensembl_id']),axis=1).stack().reset_index(level=1, drop=True)
+        ensembl = self.df.apply(
+            lambda x: pd.Series(x['ensembl_id'], dtype='object'),
+            axis=1
+        ).stack().reset_index(level=1, drop=True)
         ensembl.name = 'ensembl_id'
 
-        return self.df.drop('ensembl_id', axis=1).join(ensembl).reset_index(drop=True)[["protein_id", "protein_name", "ensembl_id", "protein_mw"]]
+        return self.df.drop('ensembl_id', axis=1).join(ensembl).reset_index(drop=True)[["protein_id", "protein_name", "gene", "ensembl_id", "protein_mw"]]
 
 class mitab:
     def __init__(self, mitabfile):
@@ -75,7 +107,7 @@ class mitab:
             else:
                 return 0
 
-        df = pd.read_csv(mitabfile, sep="\t", header = None, usecols=[0,1,2,3,14])
+        df = pd.read_csv(mitabfile, sep="\t", header = None, usecols=[0,1,2,3,14], engine='c')
         df.columns = ["bait_id","prey_id","bait_id_alt","prey_id_alt","interaction_confidence"]
 
         df['bait_id_alt'] = df['bait_id_alt'].replace(np.nan,'',regex=True)
@@ -123,7 +155,7 @@ class stringdb:
         self.df = self.read(stringdbfile, uniprot)
 
     def read(self, stringdbfile, uniprot):
-        df = pd.read_csv(stringdbfile, sep=" ")
+        df = pd.read_csv(stringdbfile, sep=" ", engine='c')
 
         df[['protein1o','protein1s']] = df.protein1.str.split('.', expand=True)
         df[['protein2o','protein2s']] = df.protein2.str.split('.', expand=True)
@@ -131,13 +163,14 @@ class stringdb:
         df = df[['protein1s','protein2s','combined_score']]
         df['combined_score'] = df['combined_score'] / 1000.0
 
-        # Map protein1
+        #Map protein 1
         df = pd.merge(df, uniprot.expand(), left_on='protein1s', right_on='ensembl_id')[['protein_id','protein2s','combined_score']]
         df.columns = ["bait_id","protein2s","combined_score"]
-        # Map protein2
+        #Map protein 2
         df = pd.merge(df, uniprot.expand(), left_on='protein2s', right_on='ensembl_id')[['bait_id','protein_id','combined_score']]
         df.columns = ["bait_id","prey_id","interaction_confidence"]
 
+
         return df
 
 class bioplex:
@@ -170,9 +203,21 @@ class preppi:
 
         return df
 
+class binary:
+    def __init__(self, binaryfile):
+        self.df = self.read(binaryfile)
+
+    def read(self, binaryfile):
+        df = pd.read_csv(binaryfile, sep=" ")
+
+        df.columns = ["bait_id","prey_id"]
+        df['interaction_confidence'] = 1
+
+        return df
+
 class net:
     def __init__(self, netfile, uniprot, meta):
-        self.formats = ['mitab','stringdb','bioplex','preppi','none']
+        self.formats = ['mitab','stringdb','bioplex','preppi', 'binary','none']
         self.format = self.identify(netfile)
 
         if self.format == 'mitab':
@@ -183,6 +228,8 @@ class net:
             network = bioplex(netfile).df
         elif self.format == 'preppi':
             network = preppi(netfile).df
+        elif self.format == 'binary':
+            network = binary(netfile).df
         elif self.format == 'none':
             protein_ids = sorted(list(meta.protein_meta['protein_id'].unique()))
             full_queries = list(itertools.combinations(protein_ids, 2))
@@ -200,7 +247,7 @@ class net:
 
     def identify(self, netfile):
         if netfile == None:
-            return self.formats[4]
+            return self.formats[5]
 
         header = pd.read_csv(netfile, sep=None, nrows=1, engine='python')
 
@@ -218,15 +265,26 @@ class net:
         # MITAB 2.5, 2.6, 2.7
         elif len(header.columns) in [11, 15, 35, 36, 42]:
             return self.formats[0]
+        # Binary
+        elif len(header.columns) == 2:
+            return self.formats[4]
         else:
             sys.exit("Error: Reference network file format is not supported.")
 
+    def get_interaction_id(self, x):
+        return '__'.join(sorted([x[0], x[1]]))
+    
     def unique_interactions(self, network):
-        def get_interaction_id(x):
-            return '__'.join(sorted([x['bait_id'], x['prey_id']]))
-
-        network['interaction_id'] = network.apply(get_interaction_id, axis=1)
-
+        # Subset network to bait_id and prey_id columns
+        bait_prey = network[["bait_id", "prey_id"]]
+        
+        # Turn subset into list of lists where each sublist is of length 2 (one bait_id one prey_id) 
+        bait_prey = bait_prey.values.tolist()
+        
+        # Use multiprocessing.Pool to distribute get_interaction_id() across list of lists
+        with multiprocessing.Pool(os.cpu_count()) as p:
+            network['interaction_id'] = p.map(self.get_interaction_id, bait_prey)
+        
         network = network.groupby('interaction_id')['interaction_confidence'].max().reset_index()
         network[['bait_id','prey_id']] = network.interaction_id.str.split('__', expand=True)
         return network[['bait_id','prey_id','interaction_confidence']]
@@ -316,6 +374,7 @@ class quantification:
 
         # Read data
         mx = pd.read_csv(infile, sep=None, engine='python')
+
         df = pd.melt(mx, id_vars=[self.protein_id_col, self.peptide_id_col], value_vars=run_id_columns, var_name=self.run_id_col, value_name=self.intensity_id_col)
 
         # Organize and rename columns
@@ -323,7 +382,8 @@ class quantification:
         df.columns = ['run_id', 'protein_id', 'peptide_id', 'peptide_intensity']
 
         # run_id, condition_id and replicate_id are categorial values, peptide_intensity must be float
-        df['run_id'] = df['run_id'].apply(lambda x: os.path.basename(str(x)))
+        if sum([1 for id in df['run_id'] if not isinstance(id, str)]) > 0:
+            df['run_id'] = df['run_id'].apply(lambda x: os.path.basename(str(x)))
         df['protein_id'] = df['protein_id'].apply(str)
         df['peptide_id'] = df['peptide_id'].apply(str)
         df['peptide_intensity'] = df['peptide_intensity'].apply(float)
@@ -354,7 +414,7 @@ class quantification:
                 return x
 
         # Read data
-        df = pd.read_csv(infile, sep="\t")
+        df = pd.read_csv(infile, sep="\t", engine='c')
 
         # Exclude decoys if present
         if 'decoy' in df.columns:
@@ -365,7 +425,8 @@ class quantification:
         df.columns = ['run_id', 'protein_id', 'peptide_id', 'peptide_intensity']
 
         # run_id, condition_id and replicate_id are categorial values, peptide_intensity must be float
-        df['run_id'] = df['run_id'].apply(lambda x: os.path.basename(str(x)))
+        if sum([1 for id in df['run_id'] if not isinstance(id, str)]) > 0:
+            df['run_id'] = df['run_id'].apply(lambda x: os.path.basename(str(x)))
         df['protein_id'] = df['protein_id'].apply(str)
         df['peptide_id'] = df['peptide_id'].apply(str)
         df['peptide_intensity'] = df['peptide_intensity'].apply(float)
@@ -408,6 +469,8 @@ class normalization:
         self.plot(self.quantification_data, self.sec_data, os.path.splitext(os.path.basename(outfile))[0]+"_raw.pdf")
         # plot normalized data
         self.plot(self.df, self.sec_data, os.path.splitext(os.path.basename(outfile))[0]+"_norm.pdf")
+        # plot number of pep identifications
+        self.plot_count(self.df, self.sec_data, os.path.splitext(os.path.basename(outfile))[0]+"_count.pdf")
 
     def slide_normalize(self, padded):
         def window(seq, n=2):
@@ -423,7 +486,6 @@ class normalization:
 
         quantification_data = self.quantification_data.copy()
         quantification_data['peptide_intensity'] = np.log2(quantification_data['peptide_intensity'])
-
         quantification_list = []
 
         if padded:
@@ -433,19 +495,25 @@ class normalization:
         else:
             min_sec_id = min(self.sec_data['sec_id'])
             max_sec_id = max(self.sec_data['sec_id'])+1
-
-        for w in window(range(min_sec_id, max_sec_id),  n=self.window_size):
+        
+        windows = [w for w in window(range(min_sec_id, max_sec_id),  n=self.window_size)]
+        mx_to_normalize = []
+        for w in windows:
             runs = self.sec_data.loc[self.sec_data['sec_id'].isin(w)]['run_id']
-            mx_norm = self.normalize(quantification_data.loc[quantification_data['run_id'].isin(runs)])
-            quantification_list.append(mx_norm)
-
+            mx = quantification_data.loc[quantification_data['run_id'].isin(runs)]
+            mx_to_normalize.append(mx)
+        
+        # With multiprocessing
+        with multiprocessing.Pool(os.cpu_count()) as p:
+            quantification_list = p.map(self.normalize, mx_to_normalize)
+        
         quantification_norm = pd.concat(quantification_list)
         quantification_norm = quantification_norm.groupby(['run_id','protein_id','peptide_id'])['peptide_intensity'].mean().reset_index()
 
         quantification_norm['peptide_intensity'] = np.exp2(quantification_norm['peptide_intensity'])
-
+        
         return(quantification_norm)
-
+ 
     def normalize(self, quantification_data):
         def normalizeCyclicLoess(x, span=0.7, iterations = 3):
             n = len(x.columns)
@@ -491,6 +559,26 @@ class normalization:
             plt.clf()
             plt.close()
 
+    def plot_count(self, quantification_data, sec_data, filename):
+        if plt is None:
+            raise ImportError("Error: The matplotlib package is required to create a report.")
+        
+        quantification_count = quantification_data.groupby(['run_id'])['peptide_intensity'].count().reset_index() #edit
+        dfcount = pd.merge(quantification_count, sec_data, on='run_id') #edit
+        dfcount['sample_id'] = dfcount['condition_id'].astype(str) + '_' + dfcount['replicate_id'].astype(str) #edit
+        dfplot_count = pd.pivot_table(dfcount, values='peptide_intensity', index='sec_id', columns='sample_id').reset_index() #edit
+        
+        with PdfPages(filename) as pdf: #all new edit
+            plt.figure(figsize=(10, 5))
+            for sample in dfplot_count.drop('sec_id', axis=1):
+                plt.plot(dfplot_count['sec_id'], dfplot_count[sample], label=sample)
+            plt.legend()
+            plt.xlabel("SEC fraction")
+            plt.ylabel("number of peptides")
+            pdf.savefig()
+            plt.clf()
+            plt.close()
+
     def to_df(self):
         return self.df
 
@@ -501,7 +589,7 @@ class meta:
         self.decoy_right_sec_bins = decoy_right_sec_bins
 
         self.peptide_meta, self.protein_meta = self.generate(quantification_data, sec_data)
-
+ 
     def generate(self, quantification_data, sec_data):
         df = pd.merge(quantification_data, sec_data, on='run_id')
 
@@ -555,7 +643,7 @@ class query:
         self.decoy_subsample = decoy_subsample
         self.decoy_exclude = decoy_exclude
         self.df = self.generate_query(net_data, posnet_data, negnet_data, protein_meta_data)
-
+ 
     def generate_query(self, net_data, posnet_data, negnet_data, protein_meta_data):
         def _random_nonidentical_array(data):
             np.random.shuffle(data['bait_id'].values)
